indexing
	description: "{
      A primitive hash table with a fixed number of (zero-based) slots.
      Each slot contains a non-void linked list.
}"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2008, Amalasoft";
	date: "$Date: 2008/12/03 $";
	revision: "$Revision: 001$";
	howto: "{
    Create an instance of this class with a fixed number of slots.
	 Add, remove and access items as you would a HASH_TABLE
}"

class AEL_DS_FIXED_HASH_TABLE [G, H->HASHABLE]

inherit
	ANY

create
	make

--|========================================================================
feature {NONE} -- Creation
--|========================================================================

	make (sz: INTEGER) is
			-- Create Current with 'sz' slots
		require
			valid_size: sz > 0
		local
			i: INTEGER
			tl: like i_th_chain
		do
			create slots.make (0, sz - 1)
			number_of_slots := sz
			from i := minimum_index
			until i > maximum_index
			loop
				create tl.make
				tl.compare_objects
				slots.put (tl, i)
				i := i + 1
			end
		end

--|========================================================================
feature -- Structure
--|========================================================================

	slots: ARRAY [like i_th_chain]
			-- Slots allocated on creation

	number_of_slots: INTEGER
			-- Number of slots allocate

	--|--------------------------------------------------------------

	longest_chain_length: INTEGER is
			-- Length of longest chain in table
		do
			Result := longest_chain.count
		end

	--|--------------------------------------------------------------

	longest_chain: like i_th_chain is
			-- Longest chain in table
		local
			i: INTEGER
		do
			Result := slots.item (slots.lower)
			from i := slots.lower
			until i > slots.upper
			loop
				if slots.item (i).count > Result.count then
					Result := slots.item (i)
				end
				i := i + 1
			end
		end

	--|--------------------------------------------------------------

	i_th_chain (v: INTEGER): LINKED_LIST [like i_th_first] is
			-- Chain (linked list) at slot 'v'
		require
			valid_index: is_valid_slot_index (v)
		do
			Result := slots.item (v)
		end

	--|--------------------------------------------------------------

	i_th_first (v: INTEGER): TUPLE [G,H] is
			-- First item (if any) in chain (linked list) at slot 'v'
		require
			valid_index: is_valid_slot_index (v)
		local
			tl: like i_th_chain
		do
			tl := i_th_chain (v)
			if not tl.is_empty then
				Result := tl.first
			end
		end

	--|--------------------------------------------------------------

	hashed_index (k: HASHABLE): INTEGER is
		require
			exists: k /= Void
		local
			hc: INTEGER
		do
			hc := k.hash_code
			Result := hc \\ number_of_slots
		ensure
			valid: is_valid_slot_index (Result)
		end

--|========================================================================
feature -- Element change
--|========================================================================

	add_item (v: G; k: H; ff: BOOLEAN) is
			-- Add 'v' to Current at position determined by 'k'
			-- If an item by 'k' already exists, then if 'ff is True,
			-- then replace it, else do not replace it
		local
			tl: like i_th_chain
			was_found: BOOLEAN
			oc: CURSOR
		do
			tl := slots.item (hashed_index (k))
			oc := tl.cursor
			from tl.start
			until tl.exhausted or was_found
			loop
				if item_matches_key (tl.item, k) then
					was_found := True
					if ff then
						tl.replace ( [v, k] )
						count := count + 1
					end
				else
					tl.forth
				end
			end
			if not was_found then
				tl.extend ( [v, k] )
			count := count + 1
			end
			tl.go_to (oc)
		ensure
			added_if_new: (not old has (k)) implies count = count + 1
			found: has (k)
		end

--|========================================================================
feature -- Search and comparison
--|========================================================================

	search (k: H) is
			-- Search for item correspond to 'k'
			-- Set found_item to item if found
		local
			tl: like i_th_chain
			oc: CURSOR
		do
			found_cell := Void
			tl := slots.item (hashed_index (k))
			oc := tl.cursor
			from tl.start
			until tl.exhausted or found_cell /= Void
			loop
				if item_matches_key (tl.item, k) then
					found_cell := tl.item
				else
					tl.forth
				end
			end
			tl.go_to (oc)
		ensure
			found_if_has: has (k) implies found_cell /= Void
			not_found_if_not_has: (not has (k)) implies found_cell = Void
		end

	--|--------------------------------------------------------------

	item_matches_key (v: like i_th_first; k: H): BOOLEAN is
		do
			Result := v.item (2).is_equal (k)
		end

--|========================================================================
feature -- Validation
--|========================================================================

	is_valid_slot_index (v: INTEGER): BOOLEAN is
			-- Is 'v' a valid index into the slots array?
		do
			Result := v >= minimum_index and v <= maximum_index
		end

--|========================================================================
feature -- Measurement
--|========================================================================

	minimum_index: INTEGER is
			-- Lowest slot index
		do
			Result := slots.lower
		end

	maximum_index: INTEGER is
			-- Highest slot index
		do
			Result := slots.upper
		end

	--|--------------------------------------------------------------

	count: INTEGER
			-- Total number of items in all chains

--|========================================================================
feature -- Access
--|========================================================================

	item (k: H): G is
			-- Item, if any, that matches 'v'
		local
			tl: like i_th_chain
			oc: CURSOR
		do
			tl := slots.item (hashed_index (k))
			oc := tl.cursor
			from tl.start
			until tl.exhausted or Result /= Void
			loop
				if item_matches_key (tl.item, k) then
					Result ?= tl.item.item (1)
				else
					tl.forth
				end
			end
			tl.go_to (oc)
		end

--|========================================================================
feature -- Status
--|========================================================================

	has (k: H): BOOLEAN is
			-- Is there an item that matches 'k'
		local
			tl: like i_th_chain
		do
			tl := slots.item (hashed_index (k))
			Result := tl.there_exists (agent item_matches_key (?,k))
		end

	--|--------------------------------------------------------------

	found_item: G is
			-- Item found from last search, if any
		do
			if found_cell /= Void then
				Result ?= found_cell.item (1)
			end
		end

	--|--------------------------------------------------------------

	found_cell: like i_th_first
			-- Cell holding found item

--|========================================================================
feature -- Element change
--|========================================================================

	extend (v: G; k: H) is
			-- Add 'v' to Current at position determined by 'k'
			-- It is an exception to extend with an item that already 
			-- exists in Current
		require
			is_new: not has (k)
		local
			tl: like i_th_chain
		do
			tl := slots.item (hashed_index (k))
			tl.extend ( [v, k] )
			count := count + 1
		ensure
			added: count = old count + 1
			found: has (k)
		end

	--|--------------------------------------------------------------

	force (v: G; k: H) is
			-- Add 'v' to Current at position determined by 'k'
			-- If an item by 'k' already exists, then replace it
		do
			add_item (v, k, True)
		ensure
			added_if_new: (not old has (k)) implies count = old count + 1
			found: has (k)
		end

	--|--------------------------------------------------------------

	put (v: G; k: H) is
			-- Add 'v' to Current at position determined by 'k'
			-- If an item by 'k' already exists, then DO NOT replace it
		do
			add_item (v, k, False)
		ensure
			added_if_new: (not old has (k)) implies count = count + 1
			found: has (k)
		end

--|========================================================================
feature -- Element removal
--|========================================================================

	remove (k: H) is
			-- Remove item corresponding to 'k' (if any)
		local
			tl: like i_th_chain
			oc: CURSOR
			removed: BOOLEAN
		do
			tl := slots.item (hashed_index (k))
			oc := tl.cursor
			from tl.start
			until tl.exhausted or removed
			loop
				if item_matches_key (tl.item, k) then
					tl.remove
					removed := True
				else
					tl.forth
				end
			end
			if removed then
				count := count - 1
			end
		ensure
			removed: (old has (k)) implies count = (old count) - 1
			gone: not has (k)
		end

	--|--------------------------------------------------------------
invariant
	has_slots: slots /= Void

end -- class AEL_DS_FIXED_HASH_TABLE
