indexing
	description: "{
      A bounded priority queue whose items are accessible via hash keys
}"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2008, Amalasoft";
	date: "$Date: 2008/12/03 $";
	revision: "$Revision: 001$";
	howto: "{
    Create an instance of this class with a queue depth.
	 Add, remove and access items as you would a HASH_TABLE
}"

class AEL_DS_HASH_QUEUE [G, H->HASHABLE]

inherit
	ANY
		redefine
			out
		end

create
	make

--|========================================================================
feature {NONE} -- Creation
--|========================================================================

	make (sz: INTEGER) is
			-- Create Current with capacity 'sz'
		do
			capacity := sz
			create hash_table.make (sz)
			create queue.make
			queue.start
		end

--|========================================================================
feature -- Status
--|========================================================================

	capacity: INTEGER
			-- Maximum number of items in Current at any given time

	count: INTEGER is
			-- Number of items in Current presently
		do
			Result := queue.count
		end

	--|--------------------------------------------------------------

	is_full: BOOLEAN is
			-- Is the queue full?
		do
			Result := count = capacity
		end

	is_empty: BOOLEAN is
			-- Is the queue empty?
		do
			Result := count = 0
		end

 --|------------------------------------------------------------------------

	has (k: H): BOOLEAN is
			-- Does the queue have an item with the key 'k'?
		do
			Result := hash_table.has (k)
		end

	--|--------------------------------------------------------------

	out: STRING is
			-- String representation of Current
			-- Items in queue order from highest priority to lowest
		do
			if is_empty then
				Result := ""
			else
				create Result.make (count * queue.first.item (1).out.count)
				from queue.start
				until queue.exhausted
				loop
					Result.append (queue.item.item (1).out)
					queue.forth
					if not queue.exhausted then
						Result.append (", ")
					end
				end
			end
		end

--|========================================================================
feature -- Status setting
--|========================================================================

	grow (v: INTEGER) is
			-- Resize to hold 'v' items
		require
			not_smaller: v >= capacity
		do
			capacity := capacity.max (v)
		ensure
			grown: capacity = v
		end

--|========================================================================
feature -- Access
--|========================================================================

	item: G is
			-- Highest priority (oldest) item in queue
		do
			Result ?= queue.first.item (1)
		end

	--|--------------------------------------------------------------

	item_by_key (k: H): G is
			-- Item corresponding to the given key
		do
			Result := hash_table.item (k)
		end

	all_items: LINKED_LIST [G] is
			-- Contents in the form of a singly linked list
		do
			create Result.make
			Result.fill (hash_table)
		end

--|========================================================================
feature -- Element change
--|========================================================================

	extend (v: G; k: H) is
			-- Add item 'v' to queue if it's not already there
		require
			not_present: not has(k)
		do
			if is_full then
				prune_queue
			end
			queue.extend ( [ v, k ] )
			hash_table.extend (v, k)
		ensure
			added: has (k)
			larger_if_not_full: (not old is_full) implies count = old count + 1
		end

	--|--------------------------------------------------------------

	force (v: G; k: H) is
			-- Add item 'v' to queue if it's already there or not
			-- If already there, demote it to lowest priority
		do
			if has (k) then
				remove (k)
			end
			extend (v, k)
		ensure
			added: has (k)
			larger_if_not_full: not (old has (k) or old is_full) implies
				count = old count + 1
		end

	--|--------------------------------------------------------------

	put (v: G; k: H) is
			-- Add item 'v' to queue ONLY if it's not already there
			-- If already there, then do not replace the existing one
			-- and do not change the existing item's priority
		do
			if not has (k) then
				extend (v, k)
			end
		ensure
			added: has (k)
			larger_if_not_full: not (old has (k) or old is_full) implies
				count = old count + 1
		end

--|========================================================================
feature -- Element Removal
--|========================================================================

	remove (k: H) is
			-- Remove the item matching key 'k'
		require
			belongs: has (k)
		local
			found: BOOLEAN
		do
			from queue.start
			until queue.exhausted or found
			loop
				if queue.item.item (2).is_equal (k) then
					found := True
					queue.remove
					hash_table.remove (k)
				else
					queue.forth
				end
			end
			queue.start
		end

	--|--------------------------------------------------------------

	wipe_out is
		-- Remove all items
		do
			queue.wipe_out
			hash_table.wipe_out
		end

--|========================================================================
feature {NONE} -- Structure
--|========================================================================

	hash_table: HASH_TABLE [G, H]
	queue: LINKED_LIST [TUPLE [G,H]]

--|========================================================================
feature {NONE} -- Implementation
--|========================================================================

	prune_queue is
			-- Remove the highest priority (oldest) item in the queue,
			-- and its corresponding reference in the hash table
		require
			queue_is_full: is_full
		local
			tk: H
		do
			queue.start
			tk ?= queue.item.item (2)
			queue.remove
			queue.start
			hash_table.remove (tk)
		end

	--|--------------------------------------------------------------
invariant
	queue_exists: queue /= Void
	valid_queue_state:  queue.is_empty implies (queue.after or queue.before)
	bounded: count <= capacity
	consistent: queue.count = hash_table.count

end -- class AEL_DS_HASH_QUEUE
